#include <unistd.h>
#include <time.h>
#include <X11/Xlib.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "plugin.h"

typedef enum { CLOCK_TIME, CLOCK_DATE } CLOCK_FORMAT;

typedef struct _tagCLOCKMODULE
{
  MODULE base;

  /* member info */
  Display *pDisplay;
  Window root;
  Window panelWindow;
  int screen;
  LPFONT pFont;
  int textY;
  unsigned long backGround;
  int width;
  int height;
  CLOCK_FORMAT fmt;
  unsigned long black;
  char timeFmt[64];
  char dateFmt[64];
  char tooltipFmt[64];
  char tooltip[128];
} CLOCKMODULE,*LPCLOCKMODULE;

#define OFFSET 12

static void DrawClock(LPCLOCKMODULE pClock,int drawBackground)
{
  /* get the time */
  char buf[128];
  int len;
  time_t cur;
  struct tm *pTm;
  
  /* Erase the background */
  if (drawBackground)
  {
    XClearArea(pClock->pDisplay,pClock->base.panels[0].wnd,0,0,0,0,False);
  }

  time(&cur);
  pTm = localtime(&cur);

  if (pTm)
  {
    switch (pClock->fmt)
    {
    case CLOCK_TIME:
      len = strftime(buf,sizeof(buf),pClock->timeFmt,pTm);
      break;
    case CLOCK_DATE:
      len = strftime(buf,sizeof(buf),pClock->dateFmt,pTm);
      break;
    default:
      return;
    }
    
    pClock->pFont->SetColor(pClock->pFont,pClock->black);
    pClock->pFont->Render(pClock->pFont,pClock->base.panels[0].wnd,OFFSET/2,pClock->textY,buf,len);
  }
}

static void LoadOptions(LPCLOCKMODULE pClock,LPCONFIGITEM pOptions)
{
  LPCONFIGITEM pItem;

  /* set them to the defaults */
  strcpy(pClock->timeFmt,"%I:%M %p");
  strcpy(pClock->dateFmt,"%m/%d/%y");
  strcpy(pClock->tooltipFmt,"%A, %B %d, %Y");

  if (pOptions && pOptions->pSublist)
  {
    for (pItem = (LPCONFIGITEM)pOptions->pSublist->list.pHead;pItem;pItem = (LPCONFIGITEM)pItem->node.pNext)
    {
      if (strcmp(pItem->name,"timeFormat") == 0)
      {
	strcpy(pClock->timeFmt,pItem->value);
      }
      else if (strcmp(pItem->name,"dateFormat") == 0)
      {
	strcpy(pClock->dateFmt,pItem->value);
      }
      else if (strcmp(pItem->name,"tooltipFormat") == 0)
      {
	strcpy(pClock->tooltipFmt,pItem->value);
      }
    }
  }
}

static const char *Clock_GetTooltip(LPCLOCKMODULE pClock)
{
  time_t cur;
  struct tm *pTm;

  time(&cur);
  pTm = localtime(&cur);

  strftime(pClock->tooltip,sizeof(pClock->tooltip),pClock->tooltipFmt,pTm);
  return pClock->tooltip;
}

static void Clock_Timeout(LPCLOCKMODULE pClock)
{
  DrawClock(pClock,1);
  XFlush(pClock->pDisplay);
}

static void Clock_EventHandler(LPCLOCKMODULE pClock,XEvent *pEvent)
{
  int drawBackground;
  
  switch (pEvent->type)
  {
  case ButtonPress:
    drawBackground = 0;		/* don't draw background */
    {
      switch (pEvent->xbutton.button)
      {
      case 2:
      case 3:
	/* ok now we want to draw the background */
	drawBackground = 1;
	/* swap types */
        {
	  switch (pClock->fmt)
	  {
	  case CLOCK_TIME:
	    pClock->fmt = CLOCK_DATE;
	    break;
	  case CLOCK_DATE:
	    pClock->fmt = CLOCK_TIME;
	    break;
	  }
	} /* pass through to redraw */
      case 1:
	DrawClock(pClock,drawBackground);
	XFlush(pClock->pDisplay);
      }
    }
    break;
  case Expose:
    if (pEvent->xexpose.count == 0)
    {
      DrawClock(pClock,0);
    }
    break;
  }
}

static void Clock_Destroy(LPCLOCKMODULE pClock)
{
#ifdef DEBUG
  printf("clock.so, Destroy()\n");
#endif
  if (pClock->pFont)
  {
    pClock->pFont->Release(pClock->pFont);
  }

  free(pClock);
}

static Window CreateWindow(LPCLOCKMODULE pClock,LPLIBRARYINIT pInit)
{
  time_t cur;
  struct tm *pTm;
  char txt[128];
  int len;
  XSetWindowAttributes att;
  Window ret;
  
  memset(&att,0,sizeof(att));
  
#ifdef DEBUG
  printf("Creating clock window with a height of %i\n",pInit->height);
#endif
  /* get the largest width of either date or time format */
  time(&cur);
  pTm = localtime(&cur);
  len = strftime(txt,sizeof(txt),pClock->timeFmt,pTm);
  pClock->width = pInit->pFont->GetWidth(pInit->pFont,txt,len);
  
  len = strftime(txt,sizeof(txt),pClock->dateFmt,pTm);
  len = pInit->pFont->GetWidth(pInit->pFont,txt,len);

  if (len > pClock->width)
  {
    pClock->width = len;
  }

  pClock->height = pInit->height;
  pClock->width += OFFSET;
  
  /* now create the window */
  att.background_pixel = pInit->backgroundPixel;
  att.event_mask = pInit->baseInputMask | ButtonPressMask | ExposureMask;

  ret = XCreateWindow(pInit->pDisplay,pInit->panelWindow,0,0,pClock->width,pInit->height,0,CopyFromParent,InputOutput,CopyFromParent,CWEventMask | CWBackPixel,&att);
  if (ret)
  {
    /* setup GC */
    pInit->pFont->GetCenter(pInit->pFont,"12:34:56 AM",11,0,pInit->height,NULL,&pClock->textY);
    /* add ref to the font */
    pClock->pFont = pInit->pFont;
    pInit->pFont->AddRef(pInit->pFont);
  }
  return ret;
}

int Init(LPLIBRARYINIT pInit)
{
  LPCLOCKMODULE pClock = (LPCLOCKMODULE)malloc(sizeof(CLOCKMODULE));
  if (pClock)
  {
    memset(pClock,0,sizeof(CLOCKMODULE));

    pClock->base.Destroy = (MODULE_DESTROY)Clock_Destroy;
    pClock->base.Timeout = (MODULE_TIMEOUT)Clock_Timeout;
    pClock->base.EventHandler = (MODULE_EVENTHANDLER)Clock_EventHandler;
    pClock->base.GetTooltip = (MODULE_GETTOOLTIP)Clock_GetTooltip;
    pClock->base.numPanels = 1;

    pClock->pDisplay = pInit->pDisplay;
    pClock->root     = pInit->root;
    pClock->screen   = pInit->screen;
    pClock->panelWindow = pInit->panelWindow;
    pClock->backGround = pInit->backgroundPixel;
    pClock->fmt    = CLOCK_TIME;
    pClock->black  = BlackPixel(pInit->pDisplay,pInit->screen);

    /* load formats */
    LoadOptions(pClock,pInit->pOptions);

    pClock->base.panels[0].name = "Clock";
    if (!(pClock->base.panels[0].wnd  = CreateWindow(pClock,pInit)))
    {
      free(pClock);
      return 0;
    }

    /* set return value */
    pInit->pModule = (LPMODULE)pClock;
  }
  return (pClock != 0);
}

